"""Helpers to get documentation metadata dynamically."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, get_type_hints

from griffe_typingdoc._docstrings import _to_parameters_section

if TYPE_CHECKING:
    from griffe import Attribute, Function, ObjectNode
    from griffe.docstrings.dataclasses import (
        DocstringSectionAdmonition,
        DocstringSectionOtherParameters,
        DocstringSectionParameters,
        DocstringSectionRaises,
        DocstringSectionReceives,
        DocstringSectionReturns,
        DocstringSectionWarns,
        DocstringSectionYields,
    )


def _hints(node: ObjectNode) -> dict[str, str]:
    try:
        return get_type_hints(node.obj, include_extras=True)
    except TypeError:
        if node.parent:
            return _hints(node.parent)
    return {}


def _doc(name: str, hints: dict[str, Any]) -> str | None:
    try:
        return hints[name].__metadata__[0].documentation
    except (AttributeError, KeyError):
        return None


def _attribute_docs(attr: Attribute, *, node: ObjectNode, **kwargs: Any) -> str:  # noqa: ARG001
    return _doc(attr.name, _hints(node)) or ""


def _parameters_docs(
    func: Function,
    *,
    node: ObjectNode,
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionParameters | None:
    hints = _hints(node)
    params_doc: dict[str, dict[str, Any]] = {
        name: {"description": _doc(name, hints)} for name in hints if name != "return"
    }
    if params_doc:
        return _to_parameters_section(params_doc, func)
    return None


# FIXME: Implement this function.
def _other_parameters_docs(
    func: Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionOtherParameters | None:
    return None


# FIXME: Implement this function.
def _deprecated_docs(
    attr_or_func: Attribute | Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionAdmonition | None:
    return None


# FIXME: Implement this function.
def _raises_docs(
    attr_or_func: Attribute | Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionRaises | None:
    return None


# FIXME: Implement this function.
def _warns_docs(
    attr_or_func: Attribute | Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionWarns | None:
    return None


# FIXME: Implement this function.
def _yields_docs(
    func: Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionYields | None:
    return None


# FIXME: Implement this function.
def _receives_docs(
    func: Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionReceives | None:
    return None


# FIXME: Implement this function.
def _returns_docs(
    func: Function,  # noqa: ARG001
    *,
    node: ObjectNode,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionReturns | None:
    return None



"""Development tasks."""

from __future__ import annotations

import os
import sys
from contextlib import contextmanager
from importlib.metadata import version as pkgversion
from pathlib import Path
from typing import TYPE_CHECKING

from duty import duty, tools

if TYPE_CHECKING:
    from collections.abc import Iterator

    from duty.context import Context


PY_SRC_PATHS = (Path(_) for _ in ("src", "tests", "duties.py", "scripts"))
PY_SRC_LIST = tuple(str(_) for _ in PY_SRC_PATHS)
PY_SRC = " ".join(PY_SRC_LIST)
CI = os.environ.get("CI", "0") in {"1", "true", "yes", ""}
WINDOWS = os.name == "nt"
PTY = not WINDOWS and not CI
MULTIRUN = os.environ.get("MULTIRUN", "0") == "1"


def pyprefix(title: str) -> str:  # noqa: D103
    if MULTIRUN:
        prefix = f"(python{sys.version_info.major}.{sys.version_info.minor})"
        return f"{prefix:14}{title}"
    return title


@contextmanager
def material_insiders() -> Iterator[bool]:  # noqa: D103
    if "+insiders" in pkgversion("mkdocs-material"):
        os.environ["MATERIAL_INSIDERS"] = "true"
        try:
            yield True
        finally:
            os.environ.pop("MATERIAL_INSIDERS")
    else:
        yield False


@duty
def changelog(ctx: Context, bump: str = "") -> None:
    """Update the changelog in-place with latest commits.

    Parameters:
        bump: Bump option passed to git-changelog.
    """
    ctx.run(tools.git_changelog(bump=bump or None), title="Updating changelog")


@duty(pre=["check-quality", "check-types", "check-docs", "check-api"])
def check(ctx: Context) -> None:
    """Check it all!"""


@duty
def check_quality(ctx: Context) -> None:
    """Check the code quality."""
    ctx.run(
        tools.ruff.check(*PY_SRC_LIST, config="config/ruff.toml"),
        title=pyprefix("Checking code quality"),
    )


@duty
def check_docs(ctx: Context) -> None:
    """Check if the documentation builds correctly."""
    Path("htmlcov").mkdir(parents=True, exist_ok=True)
    Path("htmlcov/index.html").touch(exist_ok=True)
    with material_insiders():
        ctx.run(
            tools.mkdocs.build(strict=True, verbose=True),
            title=pyprefix("Building documentation"),
        )


@duty
def check_types(ctx: Context) -> None:
    """Check that the code is correctly typed."""
    os.environ["FORCE_COLOR"] = "1"
    ctx.run(
        tools.mypy(*PY_SRC_LIST, config_file="config/mypy.ini"),
        title=pyprefix("Type-checking"),
    )


@duty
def check_api(ctx: Context, *cli_args: str) -> None:
    """Check for API breaking changes."""
    ctx.run(
        tools.griffe.check("griffe_typingdoc", search=["src"], color=True).add_args(*cli_args),
        title="Checking for API breaking changes",
        nofail=True,
    )


@duty
def docs(ctx: Context, *cli_args: str, host: str = "127.0.0.1", port: int = 8000) -> None:
    """Serve the documentation (localhost:8000).

    Parameters:
        host: The host to serve the docs from.
        port: The port to serve the docs on.
    """
    with material_insiders():
        ctx.run(
            tools.mkdocs.serve(dev_addr=f"{host}:{port}").add_args(*cli_args),
            title="Serving documentation",
            capture=False,
        )


@duty
def docs_deploy(ctx: Context) -> None:
    """Deploy the documentation to GitHub pages."""
    os.environ["DEPLOY"] = "true"
    with material_insiders() as insiders:
        if not insiders:
            ctx.run(lambda: False, title="Not deploying docs without Material for MkDocs Insiders!")
        ctx.run(tools.mkdocs.gh_deploy(), title="Deploying documentation")


@duty
def format(ctx: Context) -> None:
    """Run formatting tools on the code."""
    ctx.run(
        tools.ruff.check(*PY_SRC_LIST, config="config/ruff.toml", fix_only=True, exit_zero=True),
        title="Auto-fixing code",
    )
    ctx.run(tools.ruff.format(*PY_SRC_LIST, config="config/ruff.toml"), title="Formatting code")


@duty
def build(ctx: Context) -> None:
    """Build source and wheel distributions."""
    ctx.run(
        tools.build(),
        title="Building source and wheel distributions",
        pty=PTY,
    )


@duty
def publish(ctx: Context) -> None:
    """Publish source and wheel distributions to PyPI."""
    if not Path("dist").exists():
        ctx.run("false", title="No distribution files found")
    dists = [str(dist) for dist in Path("dist").iterdir()]
    ctx.run(
        tools.twine.upload(*dists, skip_existing=True),
        title="Publishing source and wheel distributions to PyPI",
        pty=PTY,
    )


@duty(post=["build", "publish", "docs-deploy"])
def release(ctx: Context, version: str = "") -> None:
    """Release a new Python package.

    Parameters:
        version: The new version number to use.
    """
    if not (version := (version or input("> Version to release: ")).strip()):
        ctx.run("false", title="A version must be provided")
    ctx.run("git add pyproject.toml CHANGELOG.md", title="Staging files", pty=PTY)
    ctx.run(["git", "commit", "-m", f"chore: Prepare release {version}"], title="Committing changes", pty=PTY)
    ctx.run(f"git tag {version}", title="Tagging commit", pty=PTY)
    ctx.run("git push", title="Pushing commits", pty=False)
    ctx.run("git push --tags", title="Pushing tags", pty=False)


@duty(silent=True, aliases=["cov"])
def coverage(ctx: Context) -> None:
    """Report coverage as text and HTML."""
    ctx.run(tools.coverage.combine(), nofail=True)
    ctx.run(tools.coverage.report(rcfile="config/coverage.ini"), capture=False)
    ctx.run(tools.coverage.html(rcfile="config/coverage.ini"))


@duty
def test(ctx: Context, *cli_args: str, match: str = "") -> None:
    """Run the test suite.

    Parameters:
        match: A pytest expression to filter selected tests.
    """
    py_version = f"{sys.version_info.major}{sys.version_info.minor}"
    os.environ["COVERAGE_FILE"] = f".coverage.{py_version}"
    ctx.run(
        tools.pytest(
            "tests",
            config_file="config/pytest.ini",
            select=match,
            color="yes",
        ).add_args("-n", "auto", *cli_args),
        title=pyprefix("Running tests"),
    )



"""This module defines the Griffe TypingDoc extension."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from griffe import Docstring, Extension, Function, ObjectNode

from griffe_typingdoc import _dynamic, _static

if TYPE_CHECKING:
    import ast
    from typing import Annotated

    from griffe.dataclasses import Attribute, Module, Object
    from typing_extensions import Doc


class TypingDocExtension(Extension):
    """Griffe extension that reads documentation from `typing.Doc`."""

    def __init__(self) -> None:
        self._handled: set[str] = set()

    def _handle_attribute(self, attr: Attribute, /, *, node: ObjectNode | None = None) -> None:
        if attr.path in self._handled:
            return
        self._handled.add(attr.path)

        module = _dynamic if node else _static

        new_sections = (
            docstring := module._attribute_docs(attr, node=node),
            deprecated_section := module._deprecated_docs(attr, node=node),
            raises_section := module._raises_docs(attr, node=node),
            warns_section := module._warns_docs(attr, node=node),
        )

        if not any(new_sections):
            return

        if not attr.docstring:
            attr.docstring = Docstring(docstring, parent=attr)

        sections = attr.docstring.parsed

        if deprecated_section:
            sections.insert(0, deprecated_section)

        if raises_section:
            sections.append(raises_section)

        if warns_section:
            sections.append(warns_section)

    def _handle_function(self, func: Function, /, *, node: ObjectNode | None = None) -> None:
        if func.path in self._handled:
            return
        self._handled.add(func.path)

        module = _dynamic if node else _static

        new_sections = (
            deprecated_section := module._deprecated_docs(func, node=node),
            params_section := module._parameters_docs(func, node=node),
            other_params_section := module._other_parameters_docs(func, node=node),
            warns_section := module._warns_docs(func, node=node),
            raises_section := module._raises_docs(func, node=node),
            yields_section := module._yields_docs(func, node=node),
            receives_section := module._receives_docs(func, node=node),
            returns_section := module._returns_docs(func, node=node),
        )

        if not any(new_sections):
            return

        if not func.docstring:
            func.docstring = Docstring("", parent=func)

        sections = func.docstring.parsed

        if other_params_section:
            sections.insert(1, other_params_section)

        if params_section:
            sections.insert(1, params_section)

        if deprecated_section:
            sections.insert(0, deprecated_section)

        if raises_section:
            sections.append(raises_section)

        if warns_section:
            sections.append(warns_section)

        if yields_section:
            sections.append(yields_section)

        if receives_section:
            sections.append(receives_section)

        if returns_section:
            sections.append(returns_section)

    def _handle_object(self, obj: Object) -> None:
        if obj.is_alias:
            return
        if obj.is_module or obj.is_class:
            for member in obj.members.values():
                self._handle_object(member)
        elif obj.is_function:
            self._handle_function(obj)
        elif obj.is_attribute:
            self._handle_attribute(obj)

    def on_package_loaded(
        self,
        *,
        pkg: Annotated[
            Module,
            Doc("The top-level module representing a package."),
        ],
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Post-process Griffe packages recursively (non-yet handled objects only)."""
        self._handle_object(pkg)

    def on_function_instance(
        self,
        *,
        node: Annotated[
            ast.AST | ObjectNode,
            Doc("The object/AST node describing the function or its definition."),
        ],
        func: Annotated[
            Function,
            Doc("""The Griffe function just instantiated."""),
        ],
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Post-process Griffe functions to add a parameters section.

        It applies only for dynamic analysis.
        """
        if isinstance(node, ObjectNode):
            self._handle_function(func, node=node)

    def on_attribute_instance(
        self,
        *,
        node: Annotated[
            ast.AST | ObjectNode,
            Doc("The object/AST node describing the attribute or its definition."),
        ],
        attr: Annotated[
            Attribute,
            Doc("The Griffe attribute just instantiated."),
        ],
        **kwargs: Any,  # noqa: ARG002
    ) -> None:
        """Post-process Griffe attributes to create their docstring.

        It applies only for dynamic analysis.
        """
        if isinstance(node, ObjectNode):
            self._handle_attribute(attr, node=node)



# This gist shows how we could get rid of docstrings micro-syntax
# like Google-style and Numpydoc-style pseudo-standards,
# using an enhanced version of PEP 727.

# The goal is to replace the following sections:

# deprecated
# Parameters
# Other Parameters
# Raises
# Warns
# Yields
# Receives
# Returns

from __future__ import annotations

from typing import (
    Annotated,
    Generator,
    deprecated,
    Doc,
    Name,
    Raises,
    Warns,
)


# Documenting deprecations, replacing Deprecated sections:
DEPRECATED: Annotated[
    int,
    deprecated(
        """Deprecated since v2.

        Please stop using this deprecated attribute, thanks!
        """
    ),
    Doc("Showing off deprecated attributes."),
]


# For functions, maybe add the information to the return value annotation:
def deprecated1() -> Annotated[None, deprecated("Deprecated since v2.")]:
    """Showing off deprecated functions."""


# For parameters:
def deprecated2(param1: Annotated[int, deprecated("Deprecated since v2."), Doc("Description of param1.")] = 0):
    """Showing off deprecated parameters."""


# Documenting exceptions, replacing Raises sections,
# maybe add the information to the return value annotation:
def exceptions() -> (
    Annotated[
        None,
        Raises(ValueError, "When something goes wrong."),
        Raises(TypeError, "When something goes even wronger."),
    ]
):
    """Showing off raised exceptions."""


# Documenting warnings, replacing Warns sections,
# maybe add the information to the return value annotation:
def warnings() -> (
    Annotated[
        None,
        Warns(FutureWarning, "Hello users."),
        Warns(DeprecationWarning, "Hello developers."),
    ]
):
    """Showing off emitted warnings."""


# Advanced use-case: documenting multiple yielded/received/returned values:
def return_tuple() -> (
    Generator[
        tuple[
            Annotated[int, Name("python"), Doc("First element of the yielded value.")],
            Annotated[float, Name("cobra"), Doc("Second element of the yielded value.")],
        ],
        tuple[
            Annotated[int, Name("beep"), Doc("First element of the received value.")],
            Annotated[float, Name("boop"), Doc("Second element of the received value.")],
        ],
        tuple[
            Annotated[int, Name("super"), Doc("First element of the returned value.")],
            Annotated[float, Name("hyper"), Doc("Second element of the returned value.")],
        ],
    ]
):
    """Showing off tuples as yield/receive/return values."""



from __future__ import annotations

from typing import (
    Annotated,
    Generator,
    Iterator,
    NotRequired,
    TypedDict,
    Unpack,
    Doc,
)

# Documenting module/class attributes, replacing Attributes sections:
ATTRIBUTE: Annotated[
    str,
    Doc(
        """Showing off attributes.

        Supporting multiple lines.
        """
    )
]


# Documenting parameters, replacing Parameters sections:
def parameters(param1: Annotated[str, Doc("Description of param1.")] = "default"):
    """Showing off parameters."""


# Documenting other parameters (keyword arguments), replacing Other Parameters sections:
class OtherParameters(TypedDict, total=False):
    """Keyword arguments of [`simple.other_parameters`][]."""
    param1: Annotated[NotRequired[str], Doc("Description of param1.")]
    param2: Annotated[NotRequired[str], Doc("Description of param2.")]


def other_parameters(
    **kwargs: Annotated[Unpack[OtherParameters], Doc("See other parameters.")],  # noqa: ARG001
) -> None:
    """Showing off other parameters."""


# Documenting returned values, replacing Returns sections:
def return_value() -> Annotated[int, Doc("Returned integer.")]:
    """Showing off return values."""
    return 0


# Documenting yielded and received values, replacing Yields and Receives sections:
def generator() -> Generator[
    Annotated[int, Doc("Yielded integers.")],
    Annotated[int, Doc("Received integers.")],
    Annotated[int, Doc("Final returned value.")],
]:
    """Showing off generators."""


# Same thing with Iterator instead of Generator:
def iterator() -> Iterator[Annotated[int, Doc("Yielded integers.")]]:
    """Showing off iterators."""


# Advanced use-case: documenting multiple yielded/received/returned values:
def return_tuple() -> Generator[
    tuple[
        Annotated[int, Doc("First element of the yielded value.")],
        Annotated[float, Doc("Second element of the yielded value.")],
    ],
    tuple[
        Annotated[int, Doc("First element of the received value.")],
        Annotated[float, Doc("Second element of the received value.")],
    ],
    tuple[
        Annotated[int, Doc("First element of the returned value.")],
        Annotated[float, Doc("Second element of the returned value.")],
    ],
]:
    """Showing off tuples as yield/receive/return values."""



"""Griffe TypingDoc package.

Griffe extension for PEP 727 - Documentation Metadata in Typing.
"""

from __future__ import annotations

from griffe_typingdoc._extension import TypingDocExtension

__all__: list[str] = ["TypingDocExtension"]



"""Helpers to get documentation metadata statically."""

from __future__ import annotations

import inspect
from ast import literal_eval
from collections import defaultdict
from typing import TYPE_CHECKING, Any

from griffe import Expr, ExprCall, ExprSubscript, ExprTuple, ParameterKind

from griffe_typingdoc._docstrings import (
    _no_self_params,
    _to_deprecated_section,
    _to_other_parameters_section,
    _to_parameters_section,
    _to_raises_section,
    _to_receives_section,
    _to_returns_section,
    _to_warns_section,
    _to_yields_section,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from griffe import Function
    from griffe.dataclasses import Attribute
    from griffe.docstrings.dataclasses import (
        DocstringSectionAdmonition,
        DocstringSectionParameters,
        DocstringSectionRaises,
        DocstringSectionReceives,
        DocstringSectionReturns,
        DocstringSectionWarns,
        DocstringSectionYields,
    )


def _literal(value: str | Expr) -> str:
    return inspect.cleandoc(literal_eval(str(value)))


def _set_metadata_doc(metadata: dict[str, Any], data: ExprCall) -> None:
    metadata["doc"] = _literal(data.arguments[0])


def _set_metadata_deprecated(metadata: dict[str, Any], data: ExprCall) -> None:
    metadata["deprecated"] = _literal(data.arguments[0])


def _set_metadata_name(metadata: dict[str, Any], data: ExprCall) -> None:
    metadata["name"] = _literal(data.arguments[0])


def _set_metadata_raises(metadata: dict[str, Any], data: ExprCall) -> None:
    metadata["raises"].append((data.arguments[0], _literal(data.arguments[1])))


def _set_metadata_warns(metadata: dict[str, Any], data: ExprCall) -> None:
    metadata["warns"].append((data.arguments[0], _literal(data.arguments[1])))


_set_metadata_map = {
    "typing.Doc": _set_metadata_doc,
    "typing_extensions.Doc": _set_metadata_doc,
    "typing.deprecated": _set_metadata_deprecated,
    "typing_extensions.deprecated": _set_metadata_deprecated,
    "typing.Name": _set_metadata_name,
    "typing_extensions.Name": _set_metadata_name,
    "typing.Raises": _set_metadata_raises,
    "typing_extensions.Raises": _set_metadata_raises,
    "typing.Warns": _set_metadata_warns,
    "typing_extensions.Warns": _set_metadata_warns,
}


def _set_metadata(metadata: dict[str, Any], data: ExprCall) -> None:
    if data.function.canonical_path in _set_metadata_map:
        _set_metadata_map[data.function.canonical_path](metadata, data)


def _metadata(annotation: str | Expr | None) -> dict[str, Any]:
    metadata: dict[str, Any] = {"raises": [], "warns": []}
    if isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
        "typing.Annotated",
        "typing_extensions.Annotated",
    }:
        annotated_data: Sequence[str | Expr]
        if isinstance(annotation.slice, ExprTuple):
            annotation, *annotated_data = annotation.slice.elements
        else:
            annotation = annotation.slice
            annotated_data = ()
        for data in annotated_data:
            if isinstance(data, ExprCall):
                _set_metadata(metadata, data)
    return metadata


def _attribute_docs(attr: Attribute, **kwargs: Any) -> str:  # noqa: ARG001
    return _metadata(attr.annotation).get("doc", "")


def _parameters_docs(func: Function, **kwargs: Any) -> DocstringSectionParameters | None:  # noqa: ARG001
    params_data: dict[str, dict[str, Any]] = defaultdict(dict)
    for parameter in _no_self_params(func):
        stars = {ParameterKind.var_positional: "*", ParameterKind.var_keyword: "**"}.get(parameter.kind, "")  # type: ignore[arg-type]
        param_name = f"{stars}{parameter.name}"
        metadata = _metadata(parameter.annotation)
        if "deprecated" in metadata or "doc" in metadata:
            description = f"{metadata.get('deprecated', '')} {metadata.get('doc', '')}".lstrip()
            params_data[param_name]["description"] = description
            params_data[param_name]["annotation"] = parameter.annotation
    if params_data:
        return _to_parameters_section(params_data, func)
    return None


def _other_parameters_docs(func: Function, **kwargs: Any) -> DocstringSectionParameters | None:  # noqa: ARG001
    for parameter in func.parameters:
        if parameter.kind is ParameterKind.var_keyword:
            annotation = parameter.annotation
            if isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
                "typing.Annotated",
                "typing_extensions.Annotated",
            }:
                annotation = annotation.slice.elements[0]  # type: ignore[union-attr]
            if isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
                "typing.Unpack",
                "typing_extensions.Unpack",
            }:
                slice_path = annotation.slice.canonical_path  # type: ignore[union-attr]
                typed_dict = func.modules_collection[slice_path]
                params_data = {
                    attr.name: {"annotation": attr.annotation, "description": description}
                    for attr in typed_dict.members.values()
                    if (description := _metadata(attr.annotation).get("doc")) is not None
                }
                if params_data:
                    return _to_other_parameters_section(params_data)
            break
    return None


def _yields_docs(func: Function, **kwargs: Any) -> DocstringSectionYields | None:  # noqa: ARG001
    yield_annotation = None
    annotation = func.returns

    if isinstance(annotation, ExprSubscript):
        if annotation.canonical_path in {"typing.Generator", "typing_extensions.Generator"}:
            yield_annotation = annotation.slice.elements[0]  # type: ignore[union-attr]
        elif annotation.canonical_path in {"typing.Iterator", "typing_extensions.Iterator"}:
            yield_annotation = annotation.slice

    if yield_annotation:
        if isinstance(yield_annotation, ExprSubscript) and yield_annotation.is_tuple:
            yield_elements = yield_annotation.slice.elements  # type: ignore[union-attr]
        else:
            yield_elements = [yield_annotation]
        yield_data = [
            {"annotation": element, **metadata}
            for element in yield_elements
            if "doc" in (metadata := _metadata(element))
        ]
        if yield_data:
            return _to_yields_section(yield_data)

    return None


def _receives_docs(func: Function, **kwargs: Any) -> DocstringSectionReceives | None:  # noqa: ARG001
    receive_annotation = None
    annotation = func.returns

    if isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
        "typing.Generator",
        "typing_extensions.Generator",
    }:
        receive_annotation = annotation.slice.elements[1]  # type: ignore[union-attr]

    if receive_annotation:
        if isinstance(receive_annotation, ExprSubscript) and receive_annotation.is_tuple:
            receive_elements = receive_annotation.slice.elements  # type: ignore[union-attr]
        else:
            receive_elements = [receive_annotation]
        receive_data = [
            {"annotation": element, **metadata}
            for element in receive_elements
            if "doc" in (metadata := _metadata(element))
        ]
        if receive_data:
            return _to_receives_section(receive_data)

    return None


def _returns_docs(func: Function, **kwargs: Any) -> DocstringSectionReturns | None:  # noqa: ARG001
    return_annotation = None
    annotation = func.returns

    if isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
        "typing.Generator",
        "typing_extensions.Generator",
    }:
        return_annotation = annotation.slice.elements[2]  # type: ignore[union-attr]
    elif isinstance(annotation, ExprSubscript) and annotation.canonical_path in {
        "typing.Annotated",
        "typing_extensions.Annotated",
    }:
        return_annotation = annotation

    if return_annotation:
        if isinstance(return_annotation, ExprSubscript) and return_annotation.is_tuple:
            return_elements = return_annotation.slice.elements  # type: ignore[union-attr]
        else:
            return_elements = [return_annotation]
        return_data = [
            {"annotation": element, **metadata}
            for element in return_elements
            if "doc" in (metadata := _metadata(element))
        ]
        if return_data:
            return _to_returns_section(return_data)

    return None


def _warns_docs(attr_or_func: Attribute | Function, **kwargs: Any) -> DocstringSectionWarns | None:  # noqa: ARG001
    if attr_or_func.is_attribute:
        annotation = attr_or_func.annotation
    elif attr_or_func.is_function:
        annotation = attr_or_func.returns
    metadata = _metadata(annotation)
    if metadata["warns"]:
        return _to_warns_section({"annotation": warned[0], "description": warned[1]} for warned in metadata["warns"])
    return None


def _raises_docs(attr_or_func: Attribute | Function, **kwargs: Any) -> DocstringSectionRaises | None:  # noqa: ARG001
    if attr_or_func.is_attribute:
        annotation = attr_or_func.annotation
    elif attr_or_func.is_function:
        annotation = attr_or_func.returns
    metadata = _metadata(annotation)
    if metadata["raises"]:
        return _to_raises_section({"annotation": raised[0], "description": raised[1]} for raised in metadata["raises"])
    return None


def _deprecated_docs(
    attr_or_func: Attribute | Function,
    **kwargs: Any,  # noqa: ARG001
) -> DocstringSectionAdmonition | None:
    if attr_or_func.is_attribute:
        annotation = attr_or_func.annotation
    elif attr_or_func.is_function:
        annotation = attr_or_func.returns
    metadata = _metadata(annotation)
    if "deprecated" in metadata:
        return _to_deprecated_section({"description": metadata["deprecated"]})
    return None



"""Tests for the Griffe extension."""

import pytest
from griffe import DocstringSectionKind, Extensions, GriffeLoader, temporary_visited_package

from griffe_typingdoc import TypingDocExtension

typing_imports = (
    "from typing import Annotated, Doc, Generator, Iterator, Name, NotRequired, Raises, TypedDict, Unpack, Warns"
)
warning_imports = "from warnings import deprecated"

# NOTE: Important! The value in calls to `Doc` will be parsed as a Name expression
# if it is valid Python syntax for names. To make sure it is correctly parsed as a string,
# it must contain invalid syntax for names, such as a dot at the end.
# The alternative solution would be to add `from __future__ import annotations`
# at the beginning of each temporary visited module.


def test_extension_on_itself() -> None:
    """Load our own package using the extension, assert a parameters section is added to the parsed docstring."""
    loader = GriffeLoader(extensions=Extensions(TypingDocExtension()))
    typingdoc = loader.load("griffe_typingdoc")
    sections = typingdoc["TypingDocExtension.on_function_instance"].docstring.parsed
    assert len(sections) == 2
    assert sections[1].kind is DocstringSectionKind.parameters
    assert sections[1].value[1].description == "The Griffe function just instantiated."


def test_attribute_doc() -> None:
    """Read documentation for attributes."""
    with temporary_visited_package(
        "package",
        modules={"__init__.py": f"{typing_imports}\na: Annotated[str, Doc('Hello.')]"},
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        assert package["a"].docstring.value == "Hello."


def test_parameter_doc() -> None:
    """Read documentation for parameters."""
    with temporary_visited_package(
        "package",
        modules={"__init__.py": f"{typing_imports}\ndef f(a: Annotated[str, Doc('Hello.')]): ..."},
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        assert package["f"].docstring.parsed[1].value[0].description == "Hello."


def test_other_parameter_doc() -> None:
    """Read documentation for other parameters, in unpack/typeddict annotations."""
    with temporary_visited_package(
        "package",
        modules={
            "__init__.py": f"""
                {typing_imports}
                class OtherParameters(TypedDict, total=False):
                    param1: Annotated[NotRequired[str], Doc("Hello.")]

                def f(**kwargs: Annotated[Unpack[OtherParameters], Doc("See other parameters.")]):
                    ...
            """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        assert package["f"].docstring.parsed[2].value[0].description == "Hello."


def test_iterator_doc() -> None:
    """Read documentation in iterator annotations."""
    with temporary_visited_package(
        "package",
        modules={
            "__init__.py": f"""
                {typing_imports}
                def f() -> Iterator[Annotated[int, Doc("Yielded hello.")]]:
                    ...
            """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        assert package["f"].docstring.parsed[1].value[0].description == "Yielded hello."


def test_generator_doc() -> None:
    """Read documentation in generator annotations."""
    with temporary_visited_package(
        "package",
        modules={
            "__init__.py": f"""
                {typing_imports}
                def f() -> Generator[
                    Annotated[int, Doc("Yielded hello.")],
                    Annotated[int, Doc("Received hello.")],
                    Annotated[int, Doc("Returned hello.")],
                ]:
                    ...
            """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert sections[1].value[0].description == "Yielded hello."
        assert sections[2].value[0].description == "Received hello."
        assert sections[3].value[0].description == "Returned hello."


def test_generator_tuples() -> None:
    """Read documentation in generator annotations (in tuples)."""
    with temporary_visited_package(
        "package",
        modules={
            "__init__.py": f"""
                {typing_imports}
                def f() -> Generator[
                    tuple[
                        Annotated[int, Doc("First yielded.")],
                        Annotated[float, Doc("Second yielded.")],
                    ],
                    tuple[
                        Annotated[int, Doc("First received.")],
                        Annotated[float, Doc("Second received.")],
                    ],
                    tuple[
                        Annotated[int, Doc("First returned.")],
                        Annotated[float, Doc("Second returned.")],
                    ],
                ]:
                    ...
            """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert sections[1].value[0].description == "First yielded."
        assert sections[1].value[1].description == "Second yielded."
        assert sections[2].value[0].description == "First received."
        assert sections[2].value[1].description == "Second received."
        assert sections[3].value[0].description == "First returned."
        assert sections[3].value[1].description == "Second returned."


def test_return_doc() -> None:
    """Read documentation for return value."""
    with temporary_visited_package(
        "package",
        modules={"__init__.py": f"{typing_imports}\ndef f() -> Annotated[int, Doc('Hello.')]: ..."},
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        assert package["f"].docstring.parsed[1].value[0].description == "Hello."


def test_unpacking_typed_dict() -> None:
    """Unpack typed dicts, resolving them to their right location."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": """
                from typing import TypedDict
                from typing_extensions import Annotated, Doc, Unpack

                from package import module

                class Options(TypedDict):
                    foo: Annotated[int, Doc("Foo's description.")]

                class A:
                    def __init__(self, **kwargs: Unpack[Options]) -> None:
                        '''Init.'''
                        self.options = kwargs

                class B:
                    def __init__(self, **kwargs: Unpack[module.Options]) -> None:
                        '''Init.'''
                        self.options = kwargs
                """,
            "module.py": """
                from typing import TypedDict
                from typing_extensions import Annotated, Doc

                class Options(TypedDict):
                    bar: Annotated[str, Doc("Bar's description.")]
                """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["A.__init__"].docstring.parsed
        assert len(sections) == 2
        assert sections[0].kind is DocstringSectionKind.text
        assert sections[1].kind is DocstringSectionKind.other_parameters
        foo = sections[1].value[0]
        assert foo.name == "foo"
        assert foo.description == "Foo's description."
        assert str(foo.annotation).startswith("Annotated[int")

        sections = package["B.__init__"].docstring.parsed
        assert len(sections) == 2
        assert sections[0].kind is DocstringSectionKind.text
        assert sections[1].kind is DocstringSectionKind.other_parameters
        bar = sections[1].value[0]
        assert bar.name == "bar"
        assert bar.description == "Bar's description."
        assert str(bar.annotation).startswith("Annotated[str")


@pytest.mark.parametrize(
    "annotation",
    ["int", "Annotated[int, '']"],
)
def test_ignore_unannotated_params(annotation: str) -> None:
    """Ignore parameters that are not annotated with `Doc`."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": f"{typing_imports}\ndef f(a: {annotation}):\n    '''Docstring.'''",
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert len(sections) == 1
        assert sections[0].kind is DocstringSectionKind.text


@pytest.mark.parametrize(
    "annotation",
    ["int", "Annotated[int, '']"],
)
def test_ignore_unannotated_other_params(annotation: str) -> None:
    """Ignore other parameters that are not annotated with `Doc`."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": f"""
            {typing_imports}
            from typing import TypedDict
            class Kwargs(TypedDict):
                a: {annotation}
            def f(**kwargs: Unpack[Kwargs]):
                '''Docstring.'''
            """,
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert len(sections) == 1
        assert sections[0].kind is DocstringSectionKind.text


@pytest.mark.parametrize(
    "annotation",
    ["int", "Annotated[int, '']"],
)
def test_ignore_unannotated_returns(annotation: str) -> None:
    """Ignore return values that are not annotated with `Doc`."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": f"{typing_imports}\ndef f() -> {annotation}:\n    '''Docstring.'''",
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert len(sections) == 1
        assert sections[0].kind is DocstringSectionKind.text


@pytest.mark.parametrize(
    "annotation",
    ["int", "Annotated[int, '']"],
)
def test_ignore_unannotated_yields(annotation: str) -> None:
    """Ignore yields that are not annotated with `Doc`."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": f"{typing_imports}\ndef f() -> Iterator[{annotation}]:\n    '''Docstring.'''",
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert len(sections) == 1
        assert sections[0].kind is DocstringSectionKind.text


@pytest.mark.parametrize(
    "annotation",
    ["int", "Annotated[int, '']"],
)
def test_ignore_unannotated_receives(annotation: str) -> None:
    """Ignore receives that are not annotated with `Doc`."""
    with temporary_visited_package(
        "package",
        {
            "__init__.py": f"{typing_imports}\ndef f() -> Generator[int, {annotation}, None]:\n    '''Docstring.'''",
        },
        extensions=Extensions(TypingDocExtension()),
    ) as package:
        sections = package["f"].docstring.parsed
        assert len(sections) == 1
        assert sections[0].kind is DocstringSectionKind.text



"""Helpers to build docstring sections."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from griffe import (
    DocstringParameter,
    DocstringRaise,
    DocstringReceive,
    DocstringReturn,
    DocstringSectionAdmonition,
    DocstringSectionOtherParameters,
    DocstringSectionParameters,
    DocstringSectionRaises,
    DocstringSectionReceives,
    DocstringSectionReturns,
    DocstringSectionWarns,
    DocstringSectionYields,
    DocstringWarn,
    DocstringYield,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from griffe import Function, Parameter


def _no_self_params(func: Function) -> list[Parameter]:
    if func.parent and func.parent.is_class and func.parameters and func.parameters[0].name in {"self", "cls"}:
        return list(func.parameters)[1:]
    return list(func.parameters)


def _to_parameters_section(params_dict: dict[str, dict[str, Any]], func: Function) -> DocstringSectionParameters:
    return DocstringSectionParameters(
        [
            DocstringParameter(
                name=param_name,
                description=param_doc["description"],
                annotation=param_doc["annotation"],
                value=func.parameters[param_name].default,  # type: ignore[arg-type]
            )
            for param_name, param_doc in params_dict.items()
        ],
    )


def _to_other_parameters_section(params_dict: dict[str, dict[str, Any]]) -> DocstringSectionOtherParameters:
    return DocstringSectionOtherParameters(
        [
            DocstringParameter(
                name=param_name,
                description=param_doc["description"],
                annotation=param_doc["annotation"],
            )
            for param_name, param_doc in params_dict.items()
        ],
    )


def _to_yields_section(yield_data: Iterable[dict[str, Any]]) -> DocstringSectionYields:
    return DocstringSectionYields(
        [
            DocstringYield(
                name=yield_dict.get("name", ""),
                description=yield_dict.get("doc", ""),
                annotation=yield_dict["annotation"],
            )
            for yield_dict in yield_data
        ],
    )


def _to_receives_section(receive_data: Iterable[dict[str, Any]]) -> DocstringSectionReceives:
    return DocstringSectionReceives(
        [
            DocstringReceive(
                name=receive_dict.get("name", ""),
                description=receive_dict.get("doc", ""),
                annotation=receive_dict["annotation"],
            )
            for receive_dict in receive_data
        ],
    )


def _to_returns_section(return_data: Iterable[dict[str, Any]]) -> DocstringSectionReturns:
    return DocstringSectionReturns(
        [
            DocstringReturn(
                name=return_dict.get("name", ""),
                description=return_dict.get("doc", ""),
                annotation=return_dict["annotation"],
            )
            for return_dict in return_data
        ],
    )


def _to_warns_section(warn_data: Iterable[dict[str, Any]]) -> DocstringSectionWarns:
    return DocstringSectionWarns(
        [
            DocstringWarn(
                annotation=warn_dict["annotation"],
                description=warn_dict.get("description", ""),
            )
            for warn_dict in warn_data
        ],
    )


def _to_raises_section(raise_data: Iterable[dict[str, Any]]) -> DocstringSectionRaises:
    return DocstringSectionRaises(
        [
            DocstringRaise(
                annotation=raise_dict["annotation"],
                description=raise_dict.get("description", ""),
            )
            for raise_dict in raise_data
        ],
    )


def _to_deprecated_section(deprecation_data: dict[str, Any]) -> DocstringSectionAdmonition:
    description = deprecation_data["description"]
    description_lines = deprecation_data["description"].split("\n")
    if len(description_lines) > 1:
        title = description_lines[0].strip()
        description = "\n".join(description_lines[1:]).strip()
    else:
        title = description
        description = ""
    return DocstringSectionAdmonition(kind="danger", title=title, text=description)
